/* Copyright (c) 2007 Sam Trenholme and Jean-Jacques Sarton
 *
 * TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This software is provided 'as is' with no guarantees of correctness or
 * fitness for purpose.
 */

#include "DwSocket.h"

/* Mararc parameters that are set in DwMararc.c */
extern dw_str *key_s[];
extern dw_str *key_d[];
extern int32_t key_n[];

/* Parameters set in DwSys.c */
extern int64_t the_time;
extern dwr_rg *rng_seed;

/* List of addresses we will bind to */
extern ip_addr_T bind_address[];
extern ip_addr_T upstream_address[];

/* List of active sockets */
extern SOCKET b_local[];
extern SOCKET *b_remote;

/* The list of pending remote connections */
extern remote_T *rem;

/* The numeric mararc parameters */
extern int maxprocs;
extern int timeout_seconds;
extern int dns_port;
extern int upstream_port;
extern int handle_overload;
extern int min_bind;
extern int num_ports;

#ifdef MINGW
/* Needed for the Windows way of making a socket non-blocking */
extern u_long dont_block;
#endif /*MINGW*/

/* The upstream server we will connect to (round robin rotated) */

/* Bind to all IP addresses we are to bind to and return the number of
 * IP addresses we got or INVALID_SOCKET on error */
int bind_all_udp() {
        int a = 0;
        int count = 0;
        for(a = 0; a < DW_MAXIPS; a++) {
                if(bind_address[a].len != 0) {
                        b_local[a] = do_bind(&bind_address[a],SOCK_DGRAM);
                        if(b_local[a] != INVALID_SOCKET) {
                                count++;
                        } else {
                                return INVALID_SOCKET; /* Bind failure */
                        }
                } else {
                        b_local[a] = INVALID_SOCKET;
                }
        }
        return count;
}

/* Create a sockaddr_in that will be bound to a given port; this is
 * used by the code that binds to a randomly chosen port */
void setup_bind(sockaddr_all_T *dns_udp, uint16_t port, int len) {
        if(dns_udp == 0) {
                return;
        }
        memset(dns_udp,0,sizeof(dns_udp));
        if(len == 4) {
                dns_udp->V4.sin_family = AF_INET;
                dns_udp->V4.sin_addr.s_addr = htonl(INADDR_ANY);
                dns_udp->V4.sin_port = htons(port);
#ifdef IPV6
        } else if(len == 16) {
                dns_udp->V6.sin6_family = AF_INET6;
                dns_udp->V6.sin6_addr = in6addr_any;
                dns_udp->V6.sin6_port = htons(port);
#endif
        } else { /* Bad ip */
                return;
        }
        return;
}

/* This tries to bind to a random port up to 10 times; should it fail
 * after 10 times, it returns a -1 */
int do_random_bind(SOCKET s, int len) {
        sockaddr_all_T dns_udp;
        int a = 0;
        int success = 0;

        for(a = 0; a < 10; a++) {
                /* Set up random source port to bind to */
                setup_bind(&dns_udp,
                           min_bind + (dwr_rng(rng_seed) & num_ports), len);
                /* Try to bind to that port */
                if(bind(s, (struct sockaddr *)&dns_udp,
                   sizeof(dns_udp)) != -1) {
                        success = 1;
                        break;
                }
        }
        if(success == 0) { /* Bind to random port failed */
                return -1;
        }
        return 1;
}

/* Create a sockaddr_in that will connect to a given address; -1 on fail;
 * this is used by the code that "connects" to the remote DNS server */
SOCKET setup_server(sockaddr_all_T *server, ip_addr_T *addr) {
        int s = INVALID_SOCKET;
        if(server == 0 || addr == 0) {
                return INVALID_SOCKET;
        }

        memset(server,0,sizeof(server));
        if ( addr->len == 4 ) {
                server->V4.sin_family = AF_INET;
                server->V4.sin_port = htons(upstream_port);
                memcpy(&(server->V4.sin_addr),addr->ip,4);
                s = socket(AF_INET,SOCK_DGRAM,0);
#ifdef IPV6
        } else if ( addr->len == 16 ) {
                server->V6.sin6_family = AF_INET6;
                server->V6.sin6_port = htons(upstream_port);
                memcpy(&(server->V6.sin6_addr),addr->ip,16);
                s = socket(AF_INET6,SOCK_DGRAM,0);
#endif
        } else {
                return INVALID_SOCKET;
        }
        return s;
}

/* Given a remote number, a C-string with the packet to send them, and
 * the length of that string, make a connection to a remote server */
void make_remote_connection(int32_t n, unsigned char *packet, int len) {
        sockaddr_all_T server;
        SOCKET s = 0;
        int32_t rnum = -1;
        ip_addr_T addr = {0,{0,0}};

        if(rem[n].socket != INVALID_SOCKET) { /* Already used (sanity check) */
                return;
        }

        /* Get a random query ID to send to the remote server */
        rnum = set_dns_qid(packet,len,dwr_rng(rng_seed));
        if(rnum == -1) {
                return;
        }

        /* Get IP of remote server and open a socket to them */
        addr = get_upstream_ip();
        s = setup_server(&server,&addr);
        if(s == INVALID_SOCKET) { /* Failed to make socket */
                return;
        }
        make_socket_nonblock(s); /* Linux kernel bug */

        /* Bind to source port; "connect" to remote server; send packet */
        if ((do_random_bind(s,addr.len) == -1) ||
            (connect(s, (struct sockaddr *)&server, sizeof(server)) == -1) ||
            (send(s,packet,len,0) < 0)) {
                closesocket(s);
                return;
        }

        /* OK, now note the pending connection */
        b_remote[n] = s;
        rem[n].socket = s;
        rem[n].remote_id = rnum;
}

/* Send a server failure back to the client.  This is a bit of a hack;
 * we just take the client's DNS packet, and send it as-is with QR set to
 * 1 (a response) and RCODE set to 2 (server failure) */
void send_server_fail(sockaddr_all_T *client,unsigned char *a, int len,
                      SOCKET sock) {
        socklen_t c_len = 0;

        if(len < 12) {
                return;
        }

        c_len = sizeof(sockaddr_all_T);

        a[2] |= 0x80; /* Set QR */
        a[3] &= 0xf0; a[3] |= 0x02; /* Set RCODE */

        sendto(sock,(void *)a,len,0,(struct sockaddr *)client, c_len);

}

/* Get and process a local DNS request */
void get_local_udp_packet(SOCKET sock) {
        unsigned char a[522];
        int len = 0;
        sockaddr_all_T client;
        socklen_t c_len = 0;
        ip_addr_T from_ip = {0,{0,0}};
        uint16_t from_port = 0;
        int32_t b = 0, local_id = -1;

        c_len = sizeof(client);
        make_socket_nonblock(sock); /* Linux bug workaround */
        len = recvfrom(sock,(void *)a,520,0,(struct sockaddr *)&client,
                       &c_len);

        from_port = get_from_ip_port(&from_ip,&client);

        if(check_ip_acl(&from_ip) != 1) { /* Drop unauthorized packets */
                return;
        }

        local_id = get_dns_qid(a, len, 1);
        if(local_id < 0) { /* Invalid remote packet */
                return;
        }

        b = find_free_remote();
        if(b == -1) { /* We're out of remote pending connections */
                if(handle_overload == 1) {
                        send_server_fail(&client,a,len,sock);
                }
                return;
        }
        /* Make the remote connection */
        if(rem[b].socket != INVALID_SOCKET) { /* Sanity check */
                return;
        }
        make_remote_connection(b,a,len);
        if(rem[b].socket == INVALID_SOCKET) { /* Sanity check */
                return;
        }
        rem[b].from_socket = sock;
        rem[b].ip = from_ip;
        rem[b].port = from_port;
        rem[b].local_id = local_id;
        rem[b].die = get_time() + (timeout_seconds << 8);
}

/* Forward a remote reply back to the client */
void forward_remote_reply(unsigned char *packet, size_t len, remote_T *r_ip) {
        sockaddr_all_T client;
        socklen_t len_inet = 0;

        if(r_ip == 0) {
                return;
        }
        if(r_ip->from_socket == -1) {
                return;
        }
        memset(&client,0,sizeof(client));
        len_inet = sizeof(client);

        if (r_ip->ip.len == 4) {
                client.V4.sin_family = AF_INET;
                client.V4.sin_port = htons(r_ip->port);
                memcpy(&client.V4.sin_addr, r_ip->ip.ip, 4);
#ifdef IPV6
        } else if(r_ip->ip.len == 16) {
                client.V6.sin6_family = AF_INET6;
                client.V6.sin6_port = htons(r_ip->port);
                memcpy(&client.V6.sin6_addr, r_ip->ip.ip, 16);
#endif
        } else {
                return;
        }
        sendto(r_ip->from_socket,(void *)packet,len,0,
               (struct sockaddr *)&client,len_inet);
}

/* Get and process a remote DNS request */
void get_remote_udp_packet(int b, SOCKET sock) {
        ssize_t count;
        unsigned char a[520];
        count = recv(sock,a,514,0);
        if(count < 12) {
                return;
        }
        if(count > 512) {
                return;
        }
        a[2] |= 0x80; /* Flag this as an answer (just in case they didn't) */

        /* Make sure the ID we got is the same as the one we originally
         * sent them */
        if(get_dns_qid(a,count,0) != rem[b].remote_id) {
                return;
        }

        /* Now make sure the ID is the same as the one they originally sent
         * us */
        set_dns_qid(a,count,rem[b].local_id);

        /* Send the answer we got to the appropriate local connection */
        forward_remote_reply(a,count,&rem[b]);

        /* Reset this pending remote connection */
        closesocket(b_remote[b]);
        b_remote[b] = -1;
        reset_rem(b);
}

